@@ -3,24 +3,24 @@
 
 #include "wimlib/types.h"
 
-struct string_set {
+struct string_list {
 	tchar **strings;
 	size_t num_strings;
 	size_t num_alloc_strings;
 };
 
-#define STRING_SET_INITIALIZER \
+#define STRING_LIST_INITIALIZER \
 	{ .strings = NULL, .num_strings = 0, .num_alloc_strings = 0, }
 
-#define STRING_SET(_strings) \
-	struct string_set _strings = STRING_SET_INITIALIZER
+#define STRING_LIST(_strings) \
+	struct string_list _strings = STRING_LIST_INITIALIZER
 
 typedef int (*line_mangle_t)(tchar *line, const tchar *filename,
 			     unsigned long line_no);
 
 struct text_file_section {
 	const tchar *name;
-	struct string_set *strings;
+	struct string_list *strings;
 };
 
 #define LOAD_TEXT_FILE_REMOVE_QUOTES 0x00000001
