@@ -716,6 +716,8 @@ static const char wincfg[] =
 "/swapfile.sys\n"
 "/System Volume Information\n"
 "/RECYCLER\n"
+"/$RECYCLE.BIN\n"
+"/$Recycle.Bin\n"
 "/Windows/CSC\n";
 
 static const tchar *wimboot_cfgfile =
@@ -886,6 +888,7 @@ execute_add_command(struct update_command_journal *j,
 out_destroy_config:
 	destroy_capture_config(&config);
 out:
+	FREE(params.cur_path);
 	return ret;
 }
 
